@@ -14,22 +14,27 @@
 # limitations under the License.
 
 import hashlib
-import httplib
 import os
 import random
 import socket
-import StringIO
 import time
 import urllib
 
 import simplejson as json
-
 from nose import SkipTest
 from xml.dom import minidom
+
+import six
+from six.moves import http_client
 from swiftclient import get_auth
 
+from swift.common import constraints
+from swift.common.utils import config_true_value
+
 from test import safe_repr
 
+http_client._MAXHEADERS = constraints.MAX_HEADER_COUNT
+
 
 class AuthenticationFailed(Exception):
     pass
@@ -65,7 +70,7 @@ class ResponseError(Exception):
 
 
 def listing_empty(method):
-    for i in xrange(6):
+    for i in range(6):
         if len(method()) == 0:
             return True
 
@@ -109,6 +114,7 @@ class Connection(object):
         self.auth_host = config['auth_host']
         self.auth_port = int(config['auth_port'])
         self.auth_ssl = config['auth_ssl'] in ('on', 'true', 'yes', '1')
+        self.insecure = config_true_value(config.get('insecure', 'false'))
         self.auth_prefix = config.get('auth_prefix', '/')
         self.auth_version = str(config.get('auth_version', '1'))
 
@@ -147,10 +153,11 @@ class Connection(object):
         auth_netloc = "%s:%d" % (self.auth_host, self.auth_port)
         auth_url = auth_scheme + auth_netloc + auth_path
 
+        authargs = dict(snet=False, tenant_name=self.account,
+                        auth_version=self.auth_version, os_options={},
+                        insecure=self.insecure)
         (storage_url, storage_token) = get_auth(
-            auth_url, auth_user, self.password, snet=False,
-            tenant_name=self.account, auth_version=self.auth_version,
-            os_options={})
+            auth_url, auth_user, self.password, **authargs)
 
         if not (storage_url and storage_token):
             raise AuthenticationFailed()
@@ -158,10 +165,10 @@ class Connection(object):
         x = storage_url.split('/')
 
         if x[0] == 'http:':
-            self.conn_class = httplib.HTTPConnection
+            self.conn_class = http_client.HTTPConnection
             self.storage_port = 80
         elif x[0] == 'https:':
-            self.conn_class = httplib.HTTPSConnection
+            self.conn_class = http_client.HTTPSConnection
             self.storage_port = 443
         else:
             raise ValueError('unexpected protocol %s' % (x[0]))
@@ -176,7 +183,11 @@ class Connection(object):
         self.storage_url = str('/%s/%s' % (x[3], x[4]))
         self.account_name = str(x[4])
         self.auth_user = auth_user
-        self.storage_token = storage_token
+        # With v2 keystone, storage_token is unicode.
+        # We want it to be string otherwise this would cause
+        # troubles when doing query with already encoded
+        # non ascii characters in its headers.
+        self.storage_token = str(storage_token)
         self.user_acl = '%s:%s' % (self.account, self.username)
 
         self.http_connect()
@@ -197,7 +208,7 @@ class Connection(object):
     def http_connect(self):
         self.connection = self.conn_class(self.storage_host,
                                           port=self.storage_port)
-        #self.connection.set_debuglevel(3)
+        # self.connection.set_debuglevel(3)
 
     def make_path(self, path=None, cfg=None):
         if path is None:
@@ -225,6 +236,9 @@ class Connection(object):
         if not cfg.get('no_auth_token'):
             headers['X-Auth-Token'] = self.storage_token
 
+        if cfg.get('use_token'):
+            headers['X-Auth-Token'] = cfg.get('use_token')
+
         if isinstance(hdrs, dict):
             headers.update(hdrs)
         return headers
@@ -271,7 +285,7 @@ class Connection(object):
 
             try:
                 self.response = try_request()
-            except httplib.HTTPException as e:
+            except http_client.HTTPException as e:
                 fail_messages.append(safe_repr(e))
                 continue
 
@@ -323,9 +337,9 @@ class Connection(object):
 
         self.connection = self.conn_class(self.storage_host,
                                           port=self.storage_port)
-        #self.connection.set_debuglevel(3)
+        # self.connection.set_debuglevel(3)
         self.connection.putrequest('PUT', path)
-        for key, value in headers.iteritems():
+        for key, value in headers.items():
             self.connection.putheader(key, value)
         self.connection.endheaders()
 
@@ -476,6 +490,9 @@ class Account(Base):
 
 
 class Container(Base):
+    # policy_specified is set in __init__.py when tests are being set up.
+    policy_specified = None
+
     def __init__(self, conn, account, name):
         self.conn = conn
         self.account = str(account)
@@ -488,9 +505,23 @@ class Container(Base):
             parms = {}
         if cfg is None:
             cfg = {}
+        if self.policy_specified and 'X-Storage-Policy' not in hdrs:
+            hdrs['X-Storage-Policy'] = self.policy_specified
         return self.conn.make_request('PUT', self.path, hdrs=hdrs,
                                       parms=parms, cfg=cfg) in (201, 202)
 
+    def update_metadata(self, hdrs=None, cfg=None):
+        if hdrs is None:
+            hdrs = {}
+        if cfg is None:
+            cfg = {}
+
+        self.conn.make_request('POST', self.path, hdrs=hdrs, cfg=cfg)
+        if not 200 <= self.conn.response.status <= 299:
+            raise ResponseError(self.conn.response, 'POST',
+                                self.conn.make_path(self.path))
+        return True
+
     def delete(self, hdrs=None, parms=None):
         if hdrs is None:
             hdrs = {}
@@ -582,7 +613,10 @@ class Container(Base):
         if self.conn.response.status == 204:
             required_fields = [['bytes_used', 'x-container-bytes-used'],
                                ['object_count', 'x-container-object-count']]
-            optional_fields = [['versions', 'x-versions-location']]
+            optional_fields = [
+                ['versions', 'x-versions-location'],
+                ['tempurl_key', 'x-container-meta-temp-url-key'],
+                ['tempurl_key2', 'x-container-meta-temp-url-key-2']]
 
             return self.header_fields(required_fields, optional_fields)
 
@@ -618,6 +652,9 @@ class File(Base):
             else:
                 headers['Content-Length'] = 0
 
+        if cfg.get('use_token'):
+            headers['X-Auth-Token'] = cfg.get('use_token')
+
         if cfg.get('no_content_type'):
             pass
         elif self.content_type:
@@ -635,7 +672,7 @@ class File(Base):
         block_size = 4096
 
         if isinstance(data, str):
-            data = StringIO.StringIO(data)
+            data = six.StringIO(data)
 
         checksum = hashlib.md5()
         buff = data.read(block_size)
@@ -692,13 +729,13 @@ class File(Base):
         return self.conn.make_request('COPY', self.path, hdrs=headers,
                                       parms=parms) == 201
 
-    def delete(self, hdrs=None, parms=None):
+    def delete(self, hdrs=None, parms=None, cfg=None):
         if hdrs is None:
             hdrs = {}
         if parms is None:
             parms = {}
         if self.conn.make_request('DELETE', self.path, hdrs=hdrs,
-                                  parms=parms) != 204:
+                                  cfg=cfg, parms=parms) != 204:
 
             raise ResponseError(self.conn.response, 'DELETE',
                                 self.conn.make_path(self.path))
@@ -722,8 +759,10 @@ class File(Base):
                   ['content_type', 'content-type'],
                   ['last_modified', 'last-modified'],
                   ['etag', 'etag']]
+        optional_fields = [['x_object_manifest', 'x-object-manifest']]
 
-        header_fields = self.header_fields(fields)
+        header_fields = self.header_fields(fields,
+                                           optional_fields=optional_fields)
         header_fields['etag'] = header_fields['etag'].strip('"')
         return header_fields
 
@@ -837,7 +876,7 @@ class File(Base):
         finally:
             fobj.close()
 
-    def sync_metadata(self, metadata=None, cfg=None):
+    def sync_metadata(self, metadata=None, cfg=None, parms=None):
         if metadata is None:
             metadata = {}
         if cfg is None:
@@ -854,7 +893,8 @@ class File(Base):
                 else:
                     headers['Content-Length'] = 0
 
-            self.conn.make_request('POST', self.path, hdrs=headers, cfg=cfg)
+            self.conn.make_request('POST', self.path, hdrs=headers,
+                                   parms=parms, cfg=cfg)
 
             if self.conn.response.status not in (201, 202):
                 raise ResponseError(self.conn.response, 'POST',
@@ -907,7 +947,7 @@ class File(Base):
                 pass
             self.size = int(os.fstat(data.fileno())[6])
         else:
-            data = StringIO.StringIO(data)
+            data = six.StringIO(data)
             self.size = data.len
 
         headers = self.make_headers(cfg=cfg)
@@ -959,7 +999,7 @@ class File(Base):
         if not self.write(data, hdrs=hdrs, parms=parms, cfg=cfg):
             raise ResponseError(self.conn.response, 'PUT',
                                 self.conn.make_path(self.path))
-        self.md5 = self.compute_md5sum(StringIO.StringIO(data))
+        self.md5 = self.compute_md5sum(six.StringIO(data))
         return data
 
     def write_random_return_resp(self, size=None, hdrs=None, parms=None,
@@ -976,5 +1016,28 @@ class File(Base):
                           return_resp=True)
         if not resp:
             raise ResponseError(self.conn.response)
-        self.md5 = self.compute_md5sum(StringIO.StringIO(data))
+        self.md5 = self.compute_md5sum(six.StringIO(data))
         return resp
+
+    def post(self, hdrs=None, parms=None, cfg=None, return_resp=False):
+        if hdrs is None:
+            hdrs = {}
+        if parms is None:
+            parms = {}
+        if cfg is None:
+            cfg = {}
+
+        headers = self.make_headers(cfg=cfg)
+        headers.update(hdrs)
+
+        self.conn.make_request('POST', self.path, hdrs=headers,
+                               parms=parms, cfg=cfg)
+
+        if self.conn.response.status not in (201, 202):
+            raise ResponseError(self.conn.response, 'POST',
+                                self.conn.make_path(self.path))
+
+        if return_resp:
+            return self.conn.response
+
+        return True
