@@ -50,6 +50,9 @@ class FakeLogger(object):
     def info(self, msg, *args):
         self.msg = msg
 
+    def error(self, msg, *args):
+        self.msg = msg
+
     def timing_since(*args, **kwargs):
         pass
 
@@ -74,6 +77,7 @@ class FakeBroker(object):
 class FakeAccountBroker(object):
     def __init__(self, containers):
         self.containers = containers
+        self.containers_yielded = []
 
     def get_info(self):
         info = {'account': 'a',
@@ -95,15 +99,15 @@ class FakeRing(object):
     def __init__(self):
         self.nodes = [{'id': '1',
                        'ip': '10.10.10.1',
-                       'port': None,
+                       'port': 6002,
                        'device': None},
                       {'id': '2',
-                       'ip': '10.10.10.1',
-                       'port': None,
+                       'ip': '10.10.10.2',
+                       'port': 6002,
                        'device': None},
                       {'id': '3',
-                       'ip': '10.10.10.1',
-                       'port': None,
+                       'ip': '10.10.10.3',
+                       'port': 6002,
                        'device': None},
                       ]
 
@@ -138,7 +142,7 @@ cont_nodes = [{'device': 'sda1',
 @unit.patch_policies([StoragePolicy(0, 'zero', False,
                                     object_ring=unit.FakeRing()),
                       StoragePolicy(1, 'one', True,
-                                    object_ring=unit.FakeRing())])
+                                    object_ring=unit.FakeRing(replicas=4))])
 class TestReaper(unittest.TestCase):
 
     def setUp(self):
@@ -212,7 +216,7 @@ class TestReaper(unittest.TestCase):
         r.stats_objects_possibly_remaining = 0
         r.myips = myips
         if fakelogger:
-            r.logger = FakeLogger()
+            r.logger = unit.debug_logger('test-reaper')
         return r
 
     def fake_reap_account(self, *args, **kwargs):
@@ -275,30 +279,36 @@ class TestReaper(unittest.TestCase):
             'mount_check': 'false',
         }
         r = reaper.AccountReaper(conf, logger=unit.debug_logger())
-        ring = unit.FakeRing()
         mock_path = 'swift.account.reaper.direct_delete_object'
         for policy in POLICIES:
             r.reset_stats()
             with patch(mock_path) as fake_direct_delete:
-                r.reap_object('a', 'c', 'partition', cont_nodes, 'o',
-                              policy.idx)
-                for i, call_args in enumerate(
-                        fake_direct_delete.call_args_list):
-                    cnode = cont_nodes[i]
-                    host = '%(ip)s:%(port)s' % cnode
-                    device = cnode['device']
-                    headers = {
-                        'X-Container-Host': host,
-                        'X-Container-Partition': 'partition',
-                        'X-Container-Device': device,
-                        'X-Backend-Storage-Policy-Index': policy.idx
-                    }
-                    ring = r.get_object_ring(policy.idx)
-                    expected = call(ring.devs[i], 0, 'a', 'c', 'o',
-                                    headers=headers, conn_timeout=0.5,
-                                    response_timeout=10)
-                    self.assertEqual(call_args, expected)
-            self.assertEqual(r.stats_objects_deleted, 3)
+                with patch('swift.account.reaper.time') as mock_time:
+                    mock_time.return_value = 1429117638.86767
+                    r.reap_object('a', 'c', 'partition', cont_nodes, 'o',
+                                  policy.idx)
+                    mock_time.assert_called_once_with()
+                    for i, call_args in enumerate(
+                            fake_direct_delete.call_args_list):
+                        cnode = cont_nodes[i % len(cont_nodes)]
+                        host = '%(ip)s:%(port)s' % cnode
+                        device = cnode['device']
+                        headers = {
+                            'X-Container-Host': host,
+                            'X-Container-Partition': 'partition',
+                            'X-Container-Device': device,
+                            'X-Backend-Storage-Policy-Index': policy.idx,
+                            'X-Timestamp': '1429117638.86767'
+                        }
+                        ring = r.get_object_ring(policy.idx)
+                        expected = call(dict(ring.devs[i], index=i), 0,
+                                        'a', 'c', 'o',
+                                        headers=headers, conn_timeout=0.5,
+                                        response_timeout=10)
+                        self.assertEqual(call_args, expected)
+                    self.assertEqual(policy.object_ring.replicas - 1, i)
+            self.assertEqual(r.stats_objects_deleted,
+                             policy.object_ring.replicas)
 
     def test_reap_object_fail(self):
         r = self.init_reaper({}, fakelogger=True)
@@ -309,10 +319,36 @@ class TestReaper(unittest.TestCase):
                    self.fake_direct_delete_object):
             r.reap_object('a', 'c', 'partition', cont_nodes, 'o',
                           policy.idx)
-        self.assertEqual(r.stats_objects_deleted, 1)
+        # IMHO, the stat handling in the node loop of reap object is
+        # over indented, but no one has complained, so I'm not inclined
+        # to move it.  However it's worth noting we're currently keeping
+        # stats on deletes per *replica* - which is rather obvious from
+        # these tests, but this results is surprising because of some
+        # funny logic to *skip* increments on successful deletes of
+        # replicas until we have more successful responses than
+        # failures.  This means that while the first replica doesn't
+        # increment deleted because of the failure, the second one
+        # *does* get successfully deleted, but *also does not* increment
+        # the counter (!?).
+        #
+        # In the three replica case this leaves only the last deleted
+        # object incrementing the counter - in the four replica case
+        # this leaves the last two.
+        #
+        # Basically this test will always result in:
+        #   deleted == num_replicas - 2
+        self.assertEqual(r.stats_objects_deleted,
+                         policy.object_ring.replicas - 2)
         self.assertEqual(r.stats_objects_remaining, 1)
         self.assertEqual(r.stats_objects_possibly_remaining, 1)
 
+    def test_reap_object_non_exist_policy_index(self):
+        r = self.init_reaper({}, fakelogger=True)
+        r.reap_object('a', 'c', 'partition', cont_nodes, 'o', 2)
+        self.assertEqual(r.stats_objects_deleted, 0)
+        self.assertEqual(r.stats_objects_remaining, 1)
+        self.assertEqual(r.stats_objects_possibly_remaining, 0)
+
     @patch('swift.account.reaper.Ring',
            lambda *args, **kwargs: unit.FakeRing())
     def test_reap_container(self):
@@ -335,17 +371,42 @@ class TestReaper(unittest.TestCase):
                 return headers, obj_list
 
             mocks['direct_get_container'].side_effect = fake_get_container
-            r.reap_container('a', 'partition', acc_nodes, 'c')
+            with patch('swift.account.reaper.time') as mock_time:
+                mock_time.side_effect = [1429117638.86767, 1429117639.67676]
+                r.reap_container('a', 'partition', acc_nodes, 'c')
+
+            # verify calls to direct_delete_object
             mock_calls = mocks['direct_delete_object'].call_args_list
-            self.assertEqual(3, len(mock_calls))
+            self.assertEqual(policy.object_ring.replicas, len(mock_calls))
             for call_args in mock_calls:
                 _args, kwargs = call_args
                 self.assertEqual(kwargs['headers']
                                  ['X-Backend-Storage-Policy-Index'],
                                  policy.idx)
+                self.assertEqual(kwargs['headers']
+                                 ['X-Timestamp'],
+                                 '1429117638.86767')
 
-            self.assertEquals(mocks['direct_delete_container'].call_count, 3)
-        self.assertEqual(r.stats_objects_deleted, 3)
+            # verify calls to direct_delete_container
+            self.assertEqual(mocks['direct_delete_container'].call_count, 3)
+            for i, call_args in enumerate(
+                    mocks['direct_delete_container'].call_args_list):
+                anode = acc_nodes[i % len(acc_nodes)]
+                host = '%(ip)s:%(port)s' % anode
+                device = anode['device']
+                headers = {
+                    'X-Account-Host': host,
+                    'X-Account-Partition': 'partition',
+                    'X-Account-Device': device,
+                    'X-Account-Override-Deleted': 'yes',
+                    'X-Timestamp': '1429117639.67676'
+                }
+                ring = r.get_object_ring(policy.idx)
+                expected = call(dict(ring.devs[i], index=i), 0, 'a', 'c',
+                                headers=headers, conn_timeout=0.5,
+                                response_timeout=10)
+                self.assertEqual(call_args, expected)
+        self.assertEqual(r.stats_objects_deleted, policy.object_ring.replicas)
 
     def test_reap_container_get_object_fail(self):
         r = self.init_reaper({}, fakelogger=True)
@@ -363,7 +424,7 @@ class TestReaper(unittest.TestCase):
                      self.fake_reap_object)]
         with nested(*ctx):
             r.reap_container('a', 'partition', acc_nodes, 'c')
-        self.assertEqual(r.logger.inc['return_codes.4'], 1)
+        self.assertEqual(r.logger.get_increment_counts()['return_codes.4'], 1)
         self.assertEqual(r.stats_containers_deleted, 1)
 
     def test_reap_container_partial_fail(self):
@@ -382,7 +443,7 @@ class TestReaper(unittest.TestCase):
                      self.fake_reap_object)]
         with nested(*ctx):
             r.reap_container('a', 'partition', acc_nodes, 'c')
-        self.assertEqual(r.logger.inc['return_codes.4'], 2)
+        self.assertEqual(r.logger.get_increment_counts()['return_codes.4'], 2)
         self.assertEqual(r.stats_containers_possibly_remaining, 1)
 
     def test_reap_container_full_fail(self):
@@ -401,9 +462,34 @@ class TestReaper(unittest.TestCase):
                      self.fake_reap_object)]
         with nested(*ctx):
             r.reap_container('a', 'partition', acc_nodes, 'c')
-        self.assertEqual(r.logger.inc['return_codes.4'], 3)
+        self.assertEqual(r.logger.get_increment_counts()['return_codes.4'], 3)
         self.assertEqual(r.stats_containers_remaining, 1)
 
+    @patch('swift.account.reaper.Ring',
+           lambda *args, **kwargs: unit.FakeRing())
+    def test_reap_container_non_exist_policy_index(self):
+        r = self.init_reaper({}, fakelogger=True)
+        with patch.multiple('swift.account.reaper',
+                            direct_get_container=DEFAULT,
+                            direct_delete_object=DEFAULT,
+                            direct_delete_container=DEFAULT) as mocks:
+            headers = {'X-Backend-Storage-Policy-Index': 2}
+            obj_listing = [{'name': 'o'}]
+
+            def fake_get_container(*args, **kwargs):
+                try:
+                    obj = obj_listing.pop(0)
+                except IndexError:
+                    obj_list = []
+                else:
+                    obj_list = [obj]
+                return headers, obj_list
+
+            mocks['direct_get_container'].side_effect = fake_get_container
+            r.reap_container('a', 'partition', acc_nodes, 'c')
+        self.assertEqual(r.logger.get_lines_for_level('error'), [
+            'ERROR: invalid storage policy index: 2'])
+
     def fake_reap_container(self, *args, **kwargs):
         self.called_amount += 1
         self.r.stats_containers_deleted = 1
@@ -419,21 +505,26 @@ class TestReaper(unittest.TestCase):
         self.called_amount = 0
         self.r = r = self.init_reaper({}, fakelogger=True)
         r.start_time = time.time()
-        ctx = [patch('swift.account.reaper.AccountReaper.reap_container',
-                     self.fake_reap_container),
-               patch('swift.account.reaper.AccountReaper.get_account_ring',
-                     self.fake_account_ring)]
-        with nested(*ctx):
+        with patch('swift.account.reaper.AccountReaper.reap_container',
+                   self.fake_reap_container), \
+                patch('swift.account.reaper.AccountReaper.get_account_ring',
+                      self.fake_account_ring):
             nodes = r.get_account_ring().get_part_nodes()
-            self.assertTrue(r.reap_account(broker, 'partition', nodes))
+            for container_shard, node in enumerate(nodes):
+                self.assertTrue(
+                    r.reap_account(broker, 'partition', nodes,
+                                   container_shard=container_shard))
         self.assertEqual(self.called_amount, 4)
-        self.assertEqual(r.logger.msg.find('Completed pass'), 0)
-        self.assertTrue(r.logger.msg.find('1 containers deleted'))
-        self.assertTrue(r.logger.msg.find('1 objects deleted'))
-        self.assertTrue(r.logger.msg.find('1 containers remaining'))
-        self.assertTrue(r.logger.msg.find('1 objects remaining'))
-        self.assertTrue(r.logger.msg.find('1 containers possibly remaining'))
-        self.assertTrue(r.logger.msg.find('1 objects possibly remaining'))
+        info_lines = r.logger.get_lines_for_level('info')
+        self.assertEqual(len(info_lines), 6)
+        for start_line, stat_line in zip(*[iter(info_lines)] * 2):
+            self.assertEqual(start_line, 'Beginning pass on account a')
+            self.assertTrue(stat_line.find('1 containers deleted'))
+            self.assertTrue(stat_line.find('1 objects deleted'))
+            self.assertTrue(stat_line.find('1 containers remaining'))
+            self.assertTrue(stat_line.find('1 objects remaining'))
+            self.assertTrue(stat_line.find('1 containers possibly remaining'))
+            self.assertTrue(stat_line.find('1 objects possibly remaining'))
 
     def test_reap_account_no_container(self):
         broker = FakeAccountBroker(tuple())
@@ -447,7 +538,8 @@ class TestReaper(unittest.TestCase):
         with nested(*ctx):
             nodes = r.get_account_ring().get_part_nodes()
             self.assertTrue(r.reap_account(broker, 'partition', nodes))
-        self.assertEqual(r.logger.msg.find('Completed pass'), 0)
+        self.assertTrue(r.logger.get_lines_for_level(
+            'info')[-1].startswith('Completed pass'))
         self.assertEqual(self.called_amount, 0)
 
     def test_reap_device(self):
@@ -495,6 +587,67 @@ class TestReaper(unittest.TestCase):
             r.reap_device('sda1')
         self.assertEqual(self.called_amount, 0)
 
+    def test_reap_device_with_sharding(self):
+        devices = self.prepare_data_dir()
+        conf = {'devices': devices}
+        r = self.init_reaper(conf, myips=['10.10.10.2'])
+        container_shard_used = [-1]
+
+        def fake_reap_account(*args, **kwargs):
+            container_shard_used[0] = kwargs.get('container_shard')
+
+        with patch('swift.account.reaper.AccountBroker',
+                   FakeAccountBroker), \
+                patch('swift.account.reaper.AccountReaper.get_account_ring',
+                      self.fake_account_ring), \
+                patch('swift.account.reaper.AccountReaper.reap_account',
+                      fake_reap_account):
+            r.reap_device('sda1')
+        # 10.10.10.2 is second node from ring
+        self.assertEqual(container_shard_used[0], 1)
+
+    def test_reap_account_with_sharding(self):
+        devices = self.prepare_data_dir()
+        self.called_amount = 0
+        conf = {'devices': devices}
+        r = self.init_reaper(conf, myips=['10.10.10.2'])
+
+        container_reaped = [0]
+
+        def fake_list_containers_iter(self, *args):
+            for container in self.containers:
+                if container in self.containers_yielded:
+                    continue
+
+                yield container, None, None, None
+                self.containers_yielded.append(container)
+
+        def fake_reap_container(self, account, account_partition,
+                                account_nodes, container):
+            container_reaped[0] += 1
+
+        ctx = [patch('swift.account.reaper.AccountBroker',
+                     FakeAccountBroker),
+               patch('swift.account.reaper.AccountBroker.list_containers_iter',
+                     fake_list_containers_iter),
+               patch('swift.account.reaper.AccountReaper.reap_container',
+                     fake_reap_container), ]
+        fake_ring = FakeRing()
+        with nested(*ctx):
+            fake_broker = FakeAccountBroker(['c', 'd', 'e'])
+            r.reap_account(fake_broker, 10, fake_ring.nodes, 0)
+            self.assertEqual(container_reaped[0], 1)
+
+            fake_broker = FakeAccountBroker(['c', 'd', 'e'])
+            container_reaped[0] = 0
+            r.reap_account(fake_broker, 10, fake_ring.nodes, 1)
+            self.assertEqual(container_reaped[0], 2)
+
+            container_reaped[0] = 0
+            fake_broker = FakeAccountBroker(['c', 'd', 'e'])
+            r.reap_account(fake_broker, 10, fake_ring.nodes, 2)
+            self.assertEqual(container_reaped[0], 0)
+
     def test_run_once(self):
         def prepare_data_dir():
             devices_path = tempfile.mkdtemp()
