@@ -13,19 +13,20 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import json
 import unittest
 import os
 import urllib
 from contextlib import contextmanager
-import StringIO
 from hashlib import md5
 import time
 
 import mock
+import six
 
 from swift.common import direct_client
 from swift.common.exceptions import ClientException
-from swift.common.utils import json, Timestamp
+from swift.common.utils import Timestamp
 from swift.common.swob import HeaderKeyDict, RESPONSE_REASONS
 from swift.common.storage_policy import POLICIES
 
@@ -116,7 +117,7 @@ class TestDirectClient(unittest.TestCase):
         now = time.time()
         headers = direct_client.gen_headers(add_ts=True)
         self.assertEqual(headers['user-agent'], stub_user_agent)
-        self.assert_(now - 1 < Timestamp(headers['x-timestamp']) < now + 1)
+        self.assertTrue(now - 1 < Timestamp(headers['x-timestamp']) < now + 1)
         self.assertEqual(headers['x-timestamp'],
                          Timestamp(headers['x-timestamp']).internal)
         self.assertEqual(2, len(headers))
@@ -145,7 +146,7 @@ class TestDirectClient(unittest.TestCase):
                     self.assertEqual(
                         headers['x-timestamp'],
                         Timestamp(headers['x-timestamp']).internal)
-                    self.assert_(
+                    self.assertTrue(
                         now - 1 < Timestamp(headers['x-timestamp']) < now + 1)
                 self.assertEqual(expected_header_count, len(headers))
 
@@ -193,7 +194,8 @@ class TestDirectClient(unittest.TestCase):
             'status 500',
         )
         for item in expected_err_msg_parts:
-            self.assert_(item in str(err), '%r was not in "%s"' % (item, err))
+            self.assertTrue(
+                item in str(err), '%r was not in "%s"' % (item, err))
         self.assertEqual(err.http_host, self.node['ip'])
         self.assertEqual(err.http_port, self.node['port'])
         self.assertEqual(err.http_device, self.node['device'])
@@ -230,7 +232,7 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.method, 'GET')
             self.assertEqual(conn.path, self.account_path)
         self.assertEqual(err.http_status, 500)
-        self.assert_('GET' in str(err))
+        self.assertTrue('GET' in str(err))
 
     def test_direct_delete_account(self):
         node = {'ip': '1.2.3.4', 'port': '6000', 'device': 'sda'}
@@ -247,7 +249,7 @@ class TestDirectClient(unittest.TestCase):
             path = args[3]
             self.assertEqual('/sda/0/a', path)
             headers = args[4]
-            self.assert_('X-Timestamp' in headers)
+            self.assertTrue('X-Timestamp' in headers)
 
     def test_direct_head_container(self):
         headers = HeaderKeyDict(key='value')
@@ -280,7 +282,7 @@ class TestDirectClient(unittest.TestCase):
         self.assertEqual(conn.req_headers['user-agent'], self.user_agent)
         self.assertEqual(err.http_status, 503)
         self.assertEqual(err.http_headers, headers)
-        self.assert_('HEAD' in str(err))
+        self.assertTrue('HEAD' in str(err))
 
     def test_direct_head_container_deleted(self):
         important_timestamp = Timestamp(time.time()).internal
@@ -292,7 +294,7 @@ class TestDirectClient(unittest.TestCase):
                 direct_client.direct_head_container(
                     self.node, self.part, self.account, self.container)
             except Exception as err:
-                self.assert_(isinstance(err, ClientException))
+                self.assertTrue(isinstance(err, ClientException))
             else:
                 self.fail('ClientException not raised')
             self.assertEqual(conn.method, 'HEAD')
@@ -341,6 +343,19 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.container_path)
 
+    def test_direct_delete_container_with_timestamp(self):
+        # ensure timestamp is different from any that might be auto-generated
+        timestamp = Timestamp(time.time() - 100)
+        headers = {'X-Timestamp': timestamp.internal}
+        with mocked_http_conn(200) as conn:
+            direct_client.direct_delete_container(
+                self.node, self.part, self.account, self.container,
+                headers=headers)
+            self.assertEqual(conn.method, 'DELETE')
+            self.assertEqual(conn.path, self.container_path)
+            self.assertTrue('X-Timestamp' in conn.req_headers)
+            self.assertEqual(timestamp, conn.req_headers['X-Timestamp'])
+
     def test_direct_delete_container_error(self):
         with mocked_http_conn(500) as conn:
             try:
@@ -355,7 +370,7 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.path, self.container_path)
 
         self.assertEqual(err.http_status, 500)
-        self.assert_('DELETE' in str(err))
+        self.assertTrue('DELETE' in str(err))
 
     def test_direct_put_container_object(self):
         headers = {'x-foo': 'bar'}
@@ -366,7 +381,7 @@ class TestDirectClient(unittest.TestCase):
                 headers=headers)
             self.assertEqual(conn.method, 'PUT')
             self.assertEqual(conn.path, self.obj_path)
-            self.assert_('x-timestamp' in conn.req_headers)
+            self.assertTrue('x-timestamp' in conn.req_headers)
             self.assertEqual('bar', conn.req_headers.get('x-foo'))
 
         self.assertEqual(rv, None)
@@ -386,7 +401,7 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.path, self.obj_path)
 
         self.assertEqual(err.http_status, 500)
-        self.assert_('PUT' in str(err))
+        self.assertTrue('PUT' in str(err))
 
     def test_direct_delete_container_object(self):
         with mocked_http_conn(204) as conn:
@@ -412,7 +427,7 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.path, self.obj_path)
 
         self.assertEqual(err.http_status, 500)
-        self.assert_('DELETE' in str(err))
+        self.assertTrue('DELETE' in str(err))
 
     def test_direct_head_object(self):
         headers = HeaderKeyDict({'x-foo': 'bar'})
@@ -426,8 +441,8 @@ class TestDirectClient(unittest.TestCase):
 
         self.assertEqual(conn.req_headers['user-agent'], self.user_agent)
         self.assertEqual('bar', conn.req_headers.get('x-foo'))
-        self.assert_('x-timestamp' not in conn.req_headers,
-                     'x-timestamp was in HEAD request headers')
+        self.assertTrue('x-timestamp' not in conn.req_headers,
+                        'x-timestamp was in HEAD request headers')
         self.assertEqual(headers, resp)
 
     def test_direct_head_object_error(self):
@@ -444,7 +459,7 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.path, self.obj_path)
 
         self.assertEqual(err.http_status, 500)
-        self.assert_('HEAD' in str(err))
+        self.assertTrue('HEAD' in str(err))
 
     def test_direct_head_object_not_found(self):
         important_timestamp = Timestamp(time.time()).internal
@@ -466,7 +481,7 @@ class TestDirectClient(unittest.TestCase):
                          important_timestamp)
 
     def test_direct_get_object(self):
-        contents = StringIO.StringIO('123456')
+        contents = six.StringIO('123456')
 
         with mocked_http_conn(200, body=contents) as conn:
             resp_header, obj_body = direct_client.direct_get_object(
@@ -489,7 +504,7 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.path, self.obj_path)
 
         self.assertEqual(err.http_status, 500)
-        self.assert_('GET' in str(err))
+        self.assertTrue('GET' in str(err))
 
     def test_direct_post_object(self):
         headers = {'Key': 'value'}
@@ -523,10 +538,10 @@ class TestDirectClient(unittest.TestCase):
             for header in headers:
                 self.assertEqual(conn.req_headers[header], headers[header])
             self.assertEqual(conn.req_headers['user-agent'], self.user_agent)
-            self.assert_('x-timestamp' in conn.req_headers)
+            self.assertTrue('x-timestamp' in conn.req_headers)
 
         self.assertEqual(err.http_status, 500)
-        self.assert_('POST' in str(err))
+        self.assertTrue('POST' in str(err))
 
     def test_direct_delete_object(self):
         with mocked_http_conn(200) as conn:
@@ -536,6 +551,19 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.path, self.obj_path)
         self.assertEqual(resp, None)
 
+    def test_direct_delete_object_with_timestamp(self):
+        # ensure timestamp is different from any that might be auto-generated
+        timestamp = Timestamp(time.time() - 100)
+        headers = {'X-Timestamp': timestamp.internal}
+        with mocked_http_conn(200) as conn:
+            direct_client.direct_delete_object(
+                self.node, self.part, self.account, self.container, self.obj,
+                headers=headers)
+            self.assertEqual(conn.method, 'DELETE')
+            self.assertEqual(conn.path, self.obj_path)
+            self.assertTrue('X-Timestamp' in conn.req_headers)
+            self.assertEqual(timestamp, conn.req_headers['X-Timestamp'])
+
     def test_direct_delete_object_error(self):
         with mocked_http_conn(503) as conn:
             try:
@@ -549,10 +577,10 @@ class TestDirectClient(unittest.TestCase):
             self.assertEqual(conn.method, 'DELETE')
             self.assertEqual(conn.path, self.obj_path)
         self.assertEqual(err.http_status, 503)
-        self.assert_('DELETE' in str(err))
+        self.assertTrue('DELETE' in str(err))
 
     def test_direct_put_object_with_content_length(self):
-        contents = StringIO.StringIO('123456')
+        contents = six.StringIO('123456')
 
         with mocked_http_conn(200) as conn:
             resp = direct_client.direct_put_object(
@@ -563,7 +591,7 @@ class TestDirectClient(unittest.TestCase):
         self.assertEqual(md5('123456').hexdigest(), resp)
 
     def test_direct_put_object_fail(self):
-        contents = StringIO.StringIO('123456')
+        contents = six.StringIO('123456')
 
         with mocked_http_conn(500) as conn:
             try:
@@ -579,7 +607,7 @@ class TestDirectClient(unittest.TestCase):
         self.assertEqual(err.http_status, 500)
 
     def test_direct_put_object_chunked(self):
-        contents = StringIO.StringIO('123456')
+        contents = six.StringIO('123456')
 
         with mocked_http_conn(200) as conn:
             resp = direct_client.direct_put_object(
